/* grabdata.c

   grabs data from petra weather station using /dev/petra*

   (c) 1999 js <jonny@dillingen.baynet.de>
*/

#include <stdio.h>
#include <mysql.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include "petra.h"

#define TEMP_DEV		"/dev/petra0"
#define PRESSURE_DEV	"/dev/petra1"
#define HUMIDITY_DEV	"/dev/petra2"
#define WINDSPEED_DEV	"/dev/petra3"
#define WINDDIR_DEV		"/dev/petra4"
#define RAIN_DEV		"/dev/petra6"

int displayraw = 0;

void usage()
{
	printf("grabdata v0.2 - petra II weather station data grabber\n\n");
	printf("usage: grabdata [options]\n");
	printf("\t-h, --host=mysql_hostname\tspecify mysql hostname\n");
	printf("\t-d, --db=mysql_database\t\tspecify mysql database\n");
	printf("\t-t, --table=mysql_table\t\tspecify mysql table\n");
	printf("\t-u, --user=mysql_user\t\tspecify mysql user\n");
	printf("\t-p, --password=mysql_password\tspecify mysql password\n");
	printf("\t-i, --intervall=grab intervall\tpause between each grab\n");
	printf("\t-l, --loop\t\t\tloop until killed\t\n");
	printf("\t-r, --rawdata\t\t\talso display raw sensoric data\n");
	printf("\t-c, --channel\t\t\tonly display selected channel\n\n");
	printf("available channels:\n");
	printf("\t0 - temperature\n");
	printf("\t1 - pressure\n");
	printf("\t2 - humidity\n");
	printf("\t3 - windspeed\n");
	printf("\t4 - winddir\n");
	printf("\t6 - rain\n");
}

int first=0;

/* getdata

   opens the specified petra-device file and reads 2 bytes = 1 16bit value
   and returns it
   further calculations are done by the calling function

*/
unsigned int getdata(char *devname)
{
	int fd;
	unsigned int value;
	int count;

	if((fd = open(devname, O_RDONLY | O_NONBLOCK)) == -1)
	{
		fprintf(stderr, "can't open device: %s\n", devname);
		return 0;
	}

	if(!first) {
		if(ioctl(fd, PETRA_IOC_RESETRAIN) != 0) {
			perror("error in ioctl\n");
		};
		if(ioctl(fd, PETRA_IOC_STARTRAIN) != 0) {
			perror("error in ioctl\n");
		};
		first = 1;
	}

	/* read 2 bytes of information from channel */
	count = read(fd, (void*)&value, 2);

	/* we just want 16 bits */
	value &= 0x0000FFFF;

	close(fd);

	if(displayraw) {
		fprintf(stderr, "rawdata from channel %s: %d\n", devname, value);
	}

	return value;
}

int main(int argc, char *argv[])
{
	struct option long_opts[] = 
	{
		{ "user", 1, 0, 'u' },
		{ "password", 1, 0, 'p' },
		{ "intervall", 1, 0, 'i' },
		{ "host", 1, 0, 'n' },
		{ "table", 1, 0, 't' },
		{ "channel", 1, 0, 'c' },
		{ "db", 1, 0, 'd' },
		{ "rawdata", 0, 0, 'r' },
		{ "loop", 0, 0, 'l' },
		{ "help", 0, 0, 'h' }
	};
	int flag;
	MYSQL mysql;
	extern int optind;
	extern char *optarg;
	char query[1024];
	int i;
	int intervall = 1;
	char *mysql_user = "petra";
	char *mysql_pwd = "";
	char *mysql_table = "petra";
	char *mysql_db = "petra";
	char *mysql_host = "localhost";
	double temperature;
	double pressure;
	double humidity;
	double windspeed;
	double winddir;
	double rain;
	int windloops;
	int maxwindloops = 100;
	int sql = 0;
	int looprun = 0;
	int loop;
	int channel = -1;

	/* get options */
	while((i = getopt_long(argc, argv, "i:u:p:n:d:c:t:lrh", long_opts, &flag)) != EOF)
	{
		switch(i) {
			case 'n':
				mysql_host = (char *)strdup(optarg);
				sql = 1;
				break;
			case 't':
				mysql_table = (char *)strdup(optarg);
				break;
			case 'd':
				mysql_db = (char *)strdup(optarg);
				break;
			case 'u':
				mysql_user = (char *)strdup(optarg);
				break;
			case 'p':
				mysql_pwd = (char *)strdup(optarg);
				break;
			case 'l':
				looprun = 1;
				break;
			case 'i':
				intervall = atoi(optarg);
				printf("fetch data all %d seconds\n", intervall);
				break;
			case 'r':
				displayraw = 1;
				printf("raw data will be printed\n");
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'c':
				channel = atoi(optarg);
				printf("only reading channel %d\n", channel);
				break;
		}
	}

	/* all arguments must be present */
	//if (argc != 7) {
		//usage();
		//exit(1);
	//}

	if(sql)
		if(mysql_connect(&mysql, mysql_host, mysql_user, mysql_pwd) == NULL) {
			fprintf(stderr, "can't connect to db-server\n");
			exit(1);
		}

	if(sql)
		if(mysql_select_db(&mysql, mysql_db) != 0) {
			mysql_close(&mysql);
			fprintf(stderr, "can't select database on server\n");
			exit(0);
		}

	/* loop forever until killed */
	loop = 1;
	while(loop)
	{
		if(!looprun)
			loop--;

		/* read two bytes of information from each channel and write it to
		   the database
		*/
		if((channel == -1) || (channel == 0)) {
			temperature = getdata(TEMP_DEV) * 100.0 / 4095.0 - 50.0;
			printf("TEMPERATURE: %f\n", temperature);
		}

		if((channel == -1) || (channel == 1)) {
			pressure = (getdata(PRESSURE_DEV)/1000.0*180.0-46)/4.55 * 10;
			printf("PRESSURE: %f\n", pressure);
		}

		if((channel == -1) || (channel == 2)) {
			/* old formular */
			//humidity = (getdata(HUMIDITY_DEV)*100.0)/4095.0;
			/* adapted by us..., but not checked */
			humidity = 100.0 - (getdata(HUMIDITY_DEV)*100.0)/1023.0;
			printf("HUMIDITY: %f\n", humidity);
		}

		if((channel == -1) || (channel == 3)) {
			windloops = maxwindloops+1;
			windspeed = 0;
			while(--windloops) {
				windspeed += getdata(WINDSPEED_DEV);
			}
			windspeed = (40.0 * (windspeed/maxwindloops)) / 4095.0;
			printf("WINDSPEED: %f\n", windspeed);
		}
		else
		{
			windspeed = 0;
		}

		if((channel == -1) || (channel == 4)) {
			if((windspeed < 1.0) || (windspeed > 99.9))
				winddir = 0;
			else {
				/* calculate winddir */
				winddir = 360.0*getdata(WINDDIR_DEV) / 4095.0;
			}
			printf("WINDDIR: %f\n", winddir);
		}

		if((channel == -1) || (channel == 6)) {
			rain = getdata(RAIN_DEV);
			printf("rain data: %f\n", rain);
			if(rain != 5000.0) {
				rain = (255000.0 * rain) / (4095.0*31416.0);
			}
			else
				rain = 0;
			printf("RAIN: %f\n", rain);
		}

		/* insert data into sql-table */
		sprintf(query, "insert into %s values (now(), '%4.2f', '%4.2f', '%4.2f', '%4.2f', '%4.2f', '%4.2f')",
			mysql_table,
			temperature,
			windspeed,
			winddir,
			humidity,
			pressure,
			rain
		);

		if(sql)
			if(mysql_query(&mysql, query) != 0) {
				fprintf(stderr, "query failed %s\n", query);
			}

		/* sleep for a specified amount of time */
		sleep(intervall);
	}
	
	/* close database */
	if(sql)
		mysql_close(&mysql);
}
