/*
 * petra.c -- module for PETRA-II weather-station
 *
 * (c) 1999, js <jonny@bndlg.de>
 * 
 *********/

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define count_t int

#ifdef __sparc__
#  error "This module can't compile on the Sparc platform"
#else

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/ioctl.h>

#include <asm/io.h>
#include <asm/segment.h>

// #include "sysdep.h"

int petra_major = 0;
#ifdef __alpha__
int petra_base = 0x3bc; /* Hmm.... FIXME */
#else
int petra_base = 0x378; /* intel: default to lp0 */
#endif
int petra_irq = -1;
unsigned long petra_buffer = 0;
unsigned long volatile petra_head;
volatile unsigned long petra_tail;
struct wait_queue *petra_queue;

/* IOCTLs */
#define PETRA_IOC_MAGIC	'p'
#define PETRA_IOC_RESETRAIN	_IO(PETRA_IOC_MAGIC, 1)
#define PETRA_IOC_STARTRAIN _IO(PETRA_IOC_MAGIC, 2)

#define PETRA_IOC_MAXNR 2

/* to be merged into a device dependant structure..., so if you
 * have more than one petra station ...
 */
static int sel0, sel1, sel2;

/*
 * The devices with low minor numbers write/read burst of data to/from
 * specific I/O ports (by default the parallel ones).
 * 
 * The device with 128 as minor number returns ascii strings telling
 * when interrupts have been received. Writing to the device toggles
 * 00/FF on the parallel data lines. If there is a loopback wire, this
 * generates interrupts.  
 */
int petra_open (struct inode *inode, struct file *filp)
{
    MOD_INC_USE_COUNT;
    return 0;
}

void petra_release (struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
}


/* first, the port-oriented device */

enum petra_modes { PETRA_DEFAULT=0, PETRA_PAUSE, PETRA_STRING};


/* private functions */

void clock_on_off(unsigned port)
{
	outb(inb(port) ^ 0x01, port);
	udelay(3);
	outb(inb(port) ^ 0x01, port);
}

#define PETRA_START 2
#define PETRA_UNI 2
#define PETRA_BIP 255-2
#define PETRA_SINGL 2
#define PETRA_PD1 2
#define PETRA_PD0 2
#define PETRA_CLOCK_OFF 255-1
#define PETRA_CLOCK_ON	1

void setad(unsigned port)
{
	// start pulse
	outb(inb(port) | (PETRA_START), port);
	outb(inb(port) & (PETRA_CLOCK_OFF), port);
	clock_on_off(port);

	if(sel2 == 253)
		outb(inb(port) & 253, port);
	else
		outb(inb(port) | 2, port);
	clock_on_off(port);

	if(sel1 == 253)
		outb(inb(port) & 253, port);
	else
		outb(inb(port) | 2, port);
	clock_on_off(port);

	if(sel0 == 253)
		outb(inb(port) & 253, port);
	else
		outb(inb(port) | 2, port);
	clock_on_off(port);

	outb(inb(port) | PETRA_UNI, port);
	clock_on_off(port);

	outb(inb(port) | PETRA_SINGL, port);
	clock_on_off(port);

	outb(inb(port) | PETRA_PD1, port);
	clock_on_off(port);

	outb(inb(port) | PETRA_PD0, port);
	clock_on_off(port);

	outb(inb(port) & 0x01, port);
	clock_on_off(port);
}

unsigned int data_in(unsigned port)
{
	unsigned int data;
	unsigned int temp;

	data = 0;
	temp = inb(port);
	temp = inb(port) & 64;
	data = temp << 5;
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp << 4);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp << 3);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp << 2);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp << 1);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp << 0);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 1);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 2);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 3);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 4);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 5);
	clock_on_off(port);

	temp = inb(port) & 64;
	data = data + (temp >> 6);
	clock_on_off(port);

	return data;
}

void channel(int ch)
{
	switch(ch) {	
		case 0: { sel0 = 253; sel1 = 253; sel2 = 253; break; }
		case 1: { sel0 = 253; sel1 = 253; sel2 = 2; break; }
		case 2: { sel0 = 2; sel1 = 253; sel2 = 253; break; }
		case 3: { sel0 = 2; sel1 = 253; sel2 = 2; break; }
		case 4: { sel0 = 253; sel1 = 2; sel2 = 253; break; }
		case 5: { sel0 = 253; sel1 = 2; sel2 = 2; break; }
		case 6: { sel0 = 2; sel1 = 2; sel2 = 253; break; }
		case 7: { sel0 = 2; sel1 = 2; sel2 = 2; break; }
	};
}

void rain_counter_reset(unsigned port)
{	
	outb(inb(port) | 4, port);
}

void rain_counter_start(unsigned port)
{
	outb(inb(port) & 251, port);
}

unsigned int adconvert(int ch, unsigned port)
{
	channel(ch);
	setad(port);
	return data_in(port);
}

int petra_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    int retval = count;
    //unsigned port = petra_base + (MINOR(inode->i_rdev)&0x0f);
    unsigned port = petra_base;
    //int mode = (MINOR(inode->i_rdev)&0x70) >> 4;
    //unsigned char *kbuf=kmalloc(count, GFP_KERNEL), *ptr;
    unsigned char *kbuf=kmalloc(1024, GFP_KERNEL);
	unsigned ch;
	unsigned data;
   
    if (!kbuf) return -ENOMEM;

	// channel number from minor devnum
	ch = MINOR(inode->i_rdev) & 0x0f;

	//printk(KERN_INFO "reading petra channel: %x\n", ch);
	channel(ch);
	setad(port);
	data = data_in(port+1);
	//printk(KERN_INFO "data returned: %x\n", data);

	kbuf[0] = data & 0xFF;
	//printk(KERN_INFO "buf0: %x\n", kbuf[0]);
	kbuf[1] = (data >> 8) & 0xFF;
	//printk(KERN_INFO "buf1: %x\n", kbuf[1]);

	retval = 2;

	//printk(KERN_INFO "memcpy_tofs gets called...\n");

    if (retval > 0)
        memcpy_tofs(buf, kbuf, retval);
    kfree(kbuf);
    return retval;
}

int petra_ioctl(struct inode *inode, struct file *filp,
	unsigned int cmd, unsigned long arg)
{
    unsigned port = petra_base;

	/* valid cmd? */
	if(_IOC_TYPE(cmd) != PETRA_IOC_MAGIC) return -EINVAL;
	if(_IOC_NR(cmd) > PETRA_IOC_MAXNR) return -EINVAL;

	switch(cmd) {
		case PETRA_IOC_RESETRAIN:
			printk(KERN_INFO "resetting rain-counter\n");
			rain_counter_reset(port);
			break;
		case PETRA_IOC_STARTRAIN:
			printk(KERN_INFO "starting rain-counter\n");
			rain_counter_start(port);
			break;
		default:
			return -EINVAL;
	}

	return 0;
}

struct file_operations petra_fops = {
    NULL,
    petra_read,    /* petra_read */
    NULL,
    NULL,
    NULL,
    petra_ioctl,   /* petra_ioctl */
    NULL,
    petra_open,    /* petra_open */
    petra_release, /* petra_release */
    NULL,
                   /* nothing more, fill with NULLs */
};

int init_module(void)
{
    int result = check_region(petra_base,4);

    if (result) {
        printk(KERN_INFO "petra: can't get I/O address 0x%x\n", petra_base);
        return result;
    }
    request_region(petra_base,4,"petra");

    result = register_chrdev(petra_major, "petra", &petra_fops);
    if (result < 0) {
        printk(KERN_INFO "petra: can't get major number\n");
        release_region(petra_base,4);
        return result;
    }
    if (petra_major == 0) petra_major = result; /* dynamic */

    return 0;
}

void cleanup_module(void)
{
    unregister_chrdev(petra_major, "petra");
    release_region(petra_base,4);
}


#endif /* __sparc__ */
